#include <stdio.h>
#include <spu_intrinsics.h>
#include <spu_mfcio.h>


/*
 * Size of the matrices. 
 */
#define N 64


/*
 * These are the Local Store resident copies of A, B, and C
 */
volatile float A[N*N] __attribute__((aligned(16)));
volatile float B[N*N] __attribute__((aligned(16)));
volatile float C[N*N] __attribute__((aligned(16)));


/*
 * This structure is used to pass control information from
 * PPE to SPE and to pass time measurement back to PPE.
 * Needs to be a multiple of 16 bytes and correspond with the 
 * type used on the PPE on the bit-level.
 */
typedef struct
{
   uint64_t pA;
   uint64_t pB;
   uint64_t pC;
   uint32_t ticks;
   uint32_t padding;
} program_data;


/*
 * SGEMM
 */
void sgemm()
{
   /********************************************************************
    * This is a naive jik SGEMM implementation of the T, N variant     *
    * You should modify this routine so that it runs fast!             *
    ********************************************************************/
   int i, j, k;

   for( j = 0; j < N; j++ ) {
      for( i = 0; i < N; i++ ) {
         for( k = 0; k < N; k++ ) {
            C[i + j*N] += A[k + i*N] * B[k + j*N];
         }
      }
   }
}


/*
 * This is the main entry point of the SPE program 
 */
int main(unsigned long long speid, unsigned long long argp)
{
   program_data pd __attribute__((aligned(16)));
   unsigned int tag = 0;
   uint32_t t1, t2;

#define MFC_GET_AND_WAIT(ptr, ea, size, tag) \
   mfc_get(ptr, ea, size, tag, 0, 0); \
   mfc_write_tag_mask(1 << tag); \
   mfc_read_tag_status_any();

#define MFC_PUT_AND_WAIT(ptr, ea, size, tag) \
   mfc_put(ptr, ea, size, tag, 0, 0); \
   mfc_write_tag_mask(1 << tag); \
   mfc_read_tag_status_any();

   /********************************************************************
    * STAGE IN                                                         *
    ********************************************************************/
   /* Fetch program data */
   MFC_GET_AND_WAIT(&pd, argp, sizeof(program_data), tag);
   /* Fetch A */
   MFC_GET_AND_WAIT(&A[0], pd.pA, sizeof(A), tag);
   /* Fetch B */
   MFC_GET_AND_WAIT(&B[0], pd.pB, sizeof(B), tag);
   /* Fetch C */
   MFC_GET_AND_WAIT(&C[0], pd.pC, sizeof(C), tag);

   /********************************************************************
    * COMPUTE                                                          *
    ********************************************************************/
   /* Perform sgemm */
   spu_write_decrementer(0xFFFFFFFF);
   t1 = spu_read_decrementer();
   sgemm();
   t2 = spu_read_decrementer();
   pd.ticks = t1 - t2;

   /********************************************************************
    * STAGE OUT                                                        *
    ********************************************************************/
   /* Write C */
   MFC_PUT_AND_WAIT(&C[0], pd.pC, sizeof(C), tag);
   /* Write program data */
   MFC_PUT_AND_WAIT(&pd, argp, sizeof(pd), tag);

   return 0;
}
